import os
import subprocess

from rcon import Client


class VConsole():
    '''An abstract definition for the different types of console
    implementations. All implementation are intended to receive
    commands via the :any:`run` function
    '''

    def run(self, command, *params):
        '''Runs a specified command with it's parameters

        :param command: A Source Engine `console command \
        <https://developer.valvesoftware.com/wiki/Console_Command_List>`_.
        :type command: str
        :param \*params: The values to be included with the command.
        :type \*params: str
        '''
        pass

    def __enter__(self):
        pass

    def __exit__(self):
        pass


class RconConsole(VConsole):
    '''
    Issues commands by leveraging RCON.
    This is supported by most multiplayer games.
    '''

    def __init__(self, ip, port, passwd):
        '''
        :param ip: The IP the game client is listening on (usually "127.0.0.1").
        :type ip: str
        :param port: The port the game client is listening on (usually 27015).
        :type port: int
        :param passwd: The Rcon password to be used to issue commands.
        :type passwd: str'''
        self.client = Client(ip, port, passwd=passwd)

    def run(self, command, *params):
        return self.client.run(command, *params)

    def __enter__(self):
        self.client.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.__exit__(exc_type, exc_val, exc_tb)


class ExecConsole(VConsole):
    '''
    Issues commands by using the -hijack launch parameters alongside a +exec statement.
    This is supported by games that support -hijack (not csgo).
    '''

    def __init__(self, gameExe, gameDir, uuid):
        '''
        :param gameExe: The path to the game executable.
        :type gameExe: path, str
        :param gameDir: The path to the mod folder.
        :type gameDir: path, str
        :param uuid: A unique identifier for the config name.
        :type uuid: str'''

        self.gameExe = gameExe
        self.gameDir = gameDir
        self.cfgName = 'valve-exe-' + uuid + '.cfg'
        self.cfgPath = os.path.join(self.gameDir, 'cfg', self.cfgName)

    def run(self, command, *params):
        with open(self.cfgPath, "w+") as f:
            f.seek(0)
            f.write(command + ' ' + ' '.join(params))
            f.truncate()

        launch_params = [self.gameExe, '-hijack', '+exec', self.cfgName]
        self.process = subprocess.Popen(
            launch_params,
            creationflags=subprocess.DETACHED_PROCESS |
            subprocess.CREATE_NEW_PROCESS_GROUP)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            os.remove(self.cfgPath)
        except:
            pass



import os
import re
import time


class Logger():
    '''Tracks console output by leveraging the `con_logfile` command.
    Supported in most source games (except l4d2)
    '''

    def __init__(self, logpath):
        '''
        :param logpath: The path to the log file
        :type logpath: path, str
         '''
        self.logPath = logpath

        self.logs = ''  #: :type: (str) - all the logs accumulated so far
        self._bookmark = 0

    def log_ingest(self):
        '''
        Will resume reading the logs from where it last left off until the end
        and return all logs since

        :rtype: str
        '''
        logs_since_bookmark = ''
        with open(self.logPath, mode='r') as f:
            f.seek(self._bookmark, 0)
            for line in f.readlines():
                self.logs += line
                logs_since_bookmark += line
                self._bookmark = f.tell()
        return logs_since_bookmark

    def log_until(self, until=None):
        '''
        Will :any:`log_ingest()<log_ingest>` until a specified regex is
        found within the logs.

        :param until: A regex string to match against the logs
        :type until: str
         '''
        logs_since_until = ''  # all logs since the previous log_until()
        with open(self.logPath, mode='r') as f:

            while not re.search(until, logs_since_until):
                time.sleep(0.5)
                logs_since_until += self.log_ingest()
                if not until:
                    break

    def __del__(self):
        try:
            os.remove(self.logPath)
        except:
            pass



from distutils.core import setup
setup(
    name='valveexe',
    packages=['valveexe'],
    version='0.1.0',
    license='gpl-3.0',
    description='A library to interact with Source engine game\'s developer console',
    author='Maxime Dupuis',
    author_email='mdupuis@hotmail.ca',
    url='https://github.com/pySourceSDK/ValveEXE',
    download_url='https://github.com/pySourceSDK/ValveEXE/archive/v0.1.0.tar.gz',
    keywords=['exe', 'source', 'sourcesdk',
              'hammer', 'valve', 'game', 'client'],
    install_requires=['psutil', 'rcon'],
    classifiers=[
        'Development Status :: 3 - Alpha ',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.8',
    ],
)



import os
import uuid
import time
import subprocess
import psutil
import glob

from rcon import Client

from valveexe.logger import Logger
from valveexe.console import RconConsole, ExecConsole


class ValveExe(object):
    def __init__(self, gameExe, gameDir, steamExe=None, appid=None):
        '''Defines a launchable source engine game to be interacted with.

        .. note:: Some games cannot be launched by their .exe alone \
        (ex:csgo, probably for anti-cheat related reasons). \
        Those games need to include the optional parameters :any:`steamExe` \
        and :any:`appid`. Those parameters are only to be used if \
        absolutely needed, they are a fallback and will downgrade ValveEXE \
        functionnality if present.

        :param gameExe: the path for the game executable.
        :type gameExe: path, str
        :param gameDir: The mod directory.
        :type gameDir: path, str
        :param steamExe: The path for the Steam executable.
        :type steamExe: optional, path, str
        :param appid: The `Steam AppID <https://developer.valvesoftware.com/wiki/Steam_Application_IDs>`_.
        :type appid: optional, int
        '''

        self.gameExe = gameExe
        self.gameDir = gameDir
        self.exeName = self.gameExe.split('\\')[-1]

        self.appid = appid
        self.steamExe = steamExe

        self.uuid = str(uuid.uuid4())

        self.logName = 'valve-exe-' + self.uuid + '.log'
        self.logPath = os.path.join(gameDir, self.logName)

        self.console = None

        self.rcon_enabled = None
        self.hijacked = None

        self._full_cleanup()

    def launch(self, *params):
        '''Launches the game as specified in :any:`__init__` with the
        launch parameters supplied as arguments.

        :param \*params: The launch parameters to be supplied to the executable.
        :type \*params: str
        '''
        if self.steamExe and self.appid:
            self._terminate_process()  # Steam launches cannot be hijacked
            launch_params = [self.steamExe, '-applaunch', str(self.appid)]
        else:
            self.hijacked = bool(self._find_process())
            launch_params = [self.gameExe, '-hijack']
            launch_params.extend(['-game', self.gameDir])
            launch_params.extend(['+log', '0', '+sv_logflush', '1',
                                  '+con_logfile', self.logName])

        if self._check_rcon_eligible() is not False:
            launch_params.extend(['-usercon', '+ip', '0.0.0.0',
                                  '+rcon_password', self.uuid])

        launch_params.extend(list(*params))

        self.process = subprocess.Popen(
            launch_params,
            creationflags=subprocess.DETACHED_PROCESS |
            subprocess.CREATE_NEW_PROCESS_GROUP)

        while not os.path.exists(self.logPath):
            time.sleep(3)

        self.logger = Logger(self.logPath)

    def run(self, command, *params):
        '''Forwards a command with its parameters to the active :any:`VConsole`

        :param command: A Source Engine `console command \
        <https://developer.valvesoftware.com/wiki/Console_Command_List>`_.
        :type command: str
        :param \*params: The values to be included with the command.
        :type \*params: str
        '''

        if self.console:
            self.console.run(command, *params)
        else:
            with self as console:
                console.run(command, *params)

    def quit(self):
        '''Closes the game client'''
        process = self.process or self._find_process(self)
        if process:
            process.terminate()

    def _check_rcon_eligible(self):
        '''
        None: Unknown
        True: Eligible
        False: Not eligible
        '''
        process = self._find_process()
        if not process:
            # no process running
            return None
        elif self.gameDir not in process.cmdline():
            # wrong game open
            process.terminate()
            return None
        elif '-usercon' not in process.cmdline():
            # doesn't have rcon enabled
            return False
        else:
            # 'connections' confirms game is listening for rcon
            return bool(process.connections())

    def _find_process(self):
        return next((p for p in psutil.process_iter() if
                     p.name() == self.exeName), None)

    def _terminate_process(self):
        process = self._find_process()
        process and process.terminate()

    def __enter__(self):
        while self._check_rcon_eligible() is None:
            time.sleep(3)

        if self._check_rcon_eligible():
            self.console = RconConsole("127.0.0.1", 27015, self.uuid)
        else:
            self.console = ExecConsole(self.gameExe, self.gameDir, self.uuid)

        self.console.__enter__()
        return self.console

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.console.__exit__(exc_type, exc_val, exc_tb)
        self.console = None

    def __del__(self):
        try:
            self.run('con_logfile', '""')
        except:
            pass
        del self.logger

    def _full_cleanup(self):
        for f in glob.glob(self.gameDir + 'valve-exe-*.log'):
            try:
                os.remove(f)
            except:
                pass
